/*********************************************************************/
/***                                                               ***/
/***  PROJETO DE GEOMETRIA COMPUTACIONAL                           ***/
/***  Fecho Convexo - Algoritmo Embrulho para presente             ***/
/***                                                               ***/
/*** Andrea Thom Suman          NUSP 2944275                      ***/
/*** Irineu Ruiz                 NUSP 2238562                      ***/
/***                                                               ***/
/*********************************************************************/

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <stdio.h>
#include "TexFont.h"

// Constantes do Menu
#define MNU_LOAD 0
#define MNU_DELAY 1
#define MNU_SAI 2
#define MNU_EMBRULHO 3
// Tamanho da janela
#define WIDTH 500
#define HEIGTH 500

// Estrutura de um ponto bidimensional
typedef struct Ponto {
  float X;
  float Y;
} Ponto;

// Estrutara de uma cor RGB
typedef struct Cor {
  float R;
  float G;
  float B;
}Cor;

// Variveis globais que definem polgonos e fechos convexos
Ponto *poligono;
Ponto *fecho;
int len_fecho;
int npontos = 0;

// Cores definidas globalmente a serem usadas nos grficos
Cor branco;
Cor verde;
Cor azul;
Cor vermelho;

// Variveis que guardam a rotao e a translao e o delay
double zoom;
int trans_x;
int trans_y;
int delay = 1000;

// Os valores que de cada janela do programa
int main_win;
int sub_win;
int sub_ativa;

// Arquivo com a fonte a ser utilizada na janela de texto
char *filename = "default.txf";
TexFont *txf;
// Arquivo a ser careegado com os pontos do polgono
char *nome_arq = "";


// Esta funo recebe dois pontos e traa uma reta
// entre eles com a cor especificada
void tracaReta(Ponto a, Ponto b, Cor c) {
  glColor3f(c.R, c.G, c.B);
  glBegin(GL_LINES);
  glVertex2f(a.X, a.Y);
  glVertex2f(b.X, b.Y);
  glEnd();
}


// Funo que recebe um ponto e marca o mesmo na tela
void marcaPonto(Ponto a, Cor c) {
  glColor3f(c.R, c.G, c.B);
  glBegin(GL_POINTS);
  glVertex2f(a.X, a.Y);
  glEnd();
}

// Esta funo marca todos os pontos de um polgono na tela
void marcaPontos(Cor c) {
  int i; 
  for(i = 0; i < npontos; i++)
    marcaPonto(poligono[i],c);
}

// Esta funo retorna o ponto de menor Y de um polgono
//  utilizada no incio do embrulho para presente
Ponto retornaMinimo() {
  int i;
  Ponto minimo = poligono[0];
  for (i = 1; i < npontos; i++) {
    if (poligono[i].Y < minimo.Y)
      minimo = poligono[i];
    else if ((poligono[i].Y == minimo.Y) && (minimo.X < poligono[i].X))
      minimo = poligono[i];
  }
  return(minimo);
}

// Primitiva rea 2
float Area2(Ponto a, Ponto b, Ponto c) {
  return a.X * b.Y -  a.Y * b.X + a.Y * c.X - a.X * c.Y + b.X * c.Y - c.X * b.Y;
}

// Primitiva LEFT
int Left(Ponto a, Ponto b, Ponto c) {
  tracaReta(a,b,vermelho);
  tracaReta(b,c,azul);
  tracaReta(c,a,azul);      
  return (Area2(a,b,c) > 0.0);
}

// Funo que verifica se dois pontos so iguais
int equal(Ponto a, Ponto b) {
  return (a.X == b.X && a.Y == b.Y);
}

// Funo que vai demarcando o pedao j encontrado do fecho
void marcaPontosFecho(Cor c) {
  int i;
  glColor3f(c.R, c.G, c.B);
  glBegin(GL_LINE_STRIP);
  for(i = 0; i < len_fecho; i++)
    glVertex2f(fecho[i].X, fecho[i].Y);  
  glEnd();
}


// Adiciona um ponto ao fecho convexo
void addFecho(Ponto p) {
  fecho[len_fecho++] = p;
}

// A base do algoritmo de embrulho para presente
// Dado um ponto, encontra o ponto de menor ngulo em relao a si prprio
Ponto menorAngulo(Ponto p) {
  int i;
  Ponto prox;  

  for(i = 0; i < npontos; i++)
    if(!equal(poligono[i],p)) {
      prox = poligono[i];
      break;
    }

  for(i = 0; i < npontos ; i++) 
    if(!equal(poligono[i],p) && !equal(poligono[i], prox)) {
      // Reinicializando o grfico
      glClearColor(0.0, 0.0, 0.0, 1.0);
      glClear(GL_COLOR_BUFFER_BIT);
      glTranslated(trans_x, trans_y, 0.0);
      glScaled(zoom,zoom,zoom);

      // Fazendo a comparao LEFT para ver se  um ngulo menor
      if(!Left(p, prox, poligono[i])) 
	prox = poligono[i];
      
      // marcando os pontos no plano
      marcaPontos(branco);
      // Traando o fecho j encontrado
      marcaPontosFecho(verde);
      zoom = 1.0;
      trans_x = 0.0;
      trans_y = 0.0;
      glutSwapBuffers();
      // Fazendo o delay esperado
      usleep(delay);
    }
  
  return prox;
}


// O Algoritmo Embrulho para presente
void embrulho() {
  Ponto min, atual, ant;

  // Alocando espao para guardar o fecho
  len_fecho = 0;
  fecho = (Ponto *) malloc((npontos + 1) * sizeof(Ponto));  

  // Achando o ponto mnimo do polgono
  min = retornaMinimo();
  addFecho(min);
  atual = min;

  // o corpo do algoritmo, loop at completar o fecho
  do {
    ant = atual;
    atual = menorAngulo(atual); 
    addFecho(atual);
  }
  while(!equal(atual, min));  

  // Aps encerrado desenho o resultado final do algoritmo na tela
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);
  glTranslated(trans_x, trans_y, 0.0);
  glScaled(zoom,zoom,zoom);
  marcaPontos(branco);
  marcaPontosFecho(verde);
  zoom = 1.0;
  trans_x = 0.0;
  trans_y = 0.0;
  glutSwapBuffers();
}

// Funo que l os pontos de um arquivo e coloca no polgono
int lePontos(char* nome_arq) {
 FILE *entrada;
 int i;
 char c;


 entrada = fopen(nome_arq, "r");
 
 if(entrada == NULL) {
   printf("Erro na abertura do arquivo");
   return 1;
 }

 fscanf(entrada, "%d\n", &npontos);

 poligono = (Ponto *) malloc(npontos * sizeof(Ponto));
 for(i = 0; i < npontos; i++)   
   fscanf(entrada, "%f %f", &poligono[i].X, &poligono[i].Y);


 fclose(entrada);
 return 0;
}


void init(void);
void display(void);
void display_load(void);

// Funo da sub-janela que l o nome do arquivo digitado
void keyboard_load(unsigned char key, int x, int y) {
  char aux[2];
  aux[0] = key;
  aux[1] = '\0';

  if(key == 27) {
    // Se deu esc, comando abortado
    glutDestroyWindow(sub_win);
    sub_ativa = 0;
    glutSetWindow(main_win);
    glutPostRedisplay();
    strcpy(nome_arq,"\0");
  }
  else if(key == 13) {
    // Se terminou de digitar, estou destruindo a sub janela
    // e marcando os pontos
    glutDestroyWindow(sub_win);
    sub_ativa = 0;
    lePontos(nome_arq);
    glutSetWindow(main_win);
    glutPostRedisplay();
    strcpy(nome_arq,"\0");
  }
  else if(key == 8) {
    if(strlen(nome_arq) > 0) {
      nome_arq[strlen(nome_arq) - 1] = '\0';      
      glutPostRedisplay();
    }
  }
  else {
    strcat(nome_arq, aux);
    glutPostRedisplay();
  }
  
}

// Funo que abre uma sub-janela para receber o nome do arquivo
// que gerar o polgono
void load() {
   txf = txfLoadFont(filename);
  if (txf == NULL) {
    fprintf(stderr, "Problem loading %s\n", filename);
    exit(1);
  }

  sub_win = glutCreateSubWindow(main_win, 0, HEIGTH - 20, WIDTH, 300);
  sub_ativa = 1;

  glutDisplayFunc(display_load);
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(60.0, 1.0, 0.1, 20.0);
  glMatrixMode(GL_MODELVIEW);
  gluLookAt(0.0, 0.0, 5.0,
    0.0, 0.0, 0.0,
    0.0, 1.0, 0.0);

  glEnable(GL_TEXTURE_2D);
  glEnable(GL_DEPTH_TEST);

  txfEstablishTexture(txf, 0, GL_TRUE);

  glutKeyboardFunc(keyboard_load);

  glutMainLoop();
}

// Esta funo  chamada para executar os comandos de MENU do programa
void fmenu(int opt) {
  switch (opt) {
  // Foi selecionado o comando LOAD, para carregar um polgono
  case MNU_LOAD:
    load();
    break;
  // Foi chamado o comando de DELAY, que  o tempo entre uma primitiva e outra 
  case MNU_DELAY:
    break;
  // Foi chamado o Embrulho para presente
  case MNU_EMBRULHO:
    if (npontos > 0)
      embrulho();
    break;
  // Sada do programa
  case MNU_SAI:
    exit(0);
    break;
  }
}

// Funo chamada para executar alterar o delay do programa
void fdelay(int opt) {
  delay = delay + opt;
  if (delay < 0)
    delay = 0;
}

// Funo chamada para executar o zoom propriamente dito
void fzoom(int opt) {
  switch (opt) {
  case '+':
    zoom = 1.1;
    break;
  case '-':
    zoom = 0.9;
    break;
  }

  //trans_x = 0.0;
  //trans_y = 0.0;
  glutPostRedisplay();
}

// Funo chamada para executar a translao
void ftrans(int desloca) {
  switch (desloca) {
  case '>':
    trans_x = 1.0;
    trans_y = 0.0;
    break;
  case '<':
    trans_x = -1.0;
    trans_y = 0.0;
    break;
  case '^':
    trans_y = 1.0;
    trans_x = 0.0;
    break;
  case 'v':
    trans_y = -1.0;
    trans_x = 0.0;
    break;
  }
  //zoom = 1.0;
  glutPostRedisplay();
}
    
// Teclas de atalho do programa
void keyboard(unsigned char key, int x, int y) {
  if(sub_ativa) {
    glutSetWindow(sub_win);
    keyboard_load(key,x,y);
  }
  else {
    switch (key) {
    case '+':
    case '=':
      fzoom('+');
      break;
      
    case '_':
    case '-':
      fzoom('-');
      break;
    
    }
  }
}

void keyboard_special(int key, int x, int y) {
  switch (key) {
  case GLUT_KEY_RIGHT:
    ftrans('>');
    break;
    
  case GLUT_KEY_LEFT:
    ftrans('<');
    break;
    
  case GLUT_KEY_UP:
    ftrans('^');
    break;
    
  case GLUT_KEY_DOWN:
    ftrans('v');
    break;
  }  
}

// Funo MAIN do programa
int main (int argc, char **argv)
{
  int menu;
  int sub_zoom;
  int sub_trans;
  int sub_delay;

  // Inicializando as cores globais
  azul.R = 0.0;
  azul.G = 0.0;
  azul.B = 1.0;
  
  branco.R = 1.0;
  branco.G = 1.0;
  branco.B = 1.0;

  verde.R = 0.0;
  verde.G = 1.0;
  verde.B = 0.0;

  vermelho.R = 1.0;
  vermelho.G = 0.0;
  vermelho.B = 0.0;
  

  nome_arq = (char *) malloc(100*sizeof(char));

  //inicializaes do openGL
  glutInit(&argc, argv);

  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
  glutInitWindowSize(WIDTH,HEIGTH);
  main_win = glutCreateWindow("Embrulho para presente");
  
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(-25.0, 25.0, -25.0, 25.0);
  
  // Criando os menus do programa
  sub_zoom = glutCreateMenu(fzoom);
  glutAddMenuEntry("+", '+');
  glutAddMenuEntry("-", '-');
  
  sub_trans = glutCreateMenu(ftrans);
  glutAddMenuEntry(">", '>');
  glutAddMenuEntry("<", '<');
  glutAddMenuEntry("^", '^');
  glutAddMenuEntry("v", 'v');
  
  sub_delay = glutCreateMenu(fdelay);
  glutAddMenuEntry("- 100000",-100000);
  glutAddMenuEntry("- 10000",-10000);
  glutAddMenuEntry("- 1000",-1000);
  glutAddMenuEntry("- 100",-100);
  glutAddMenuEntry("+ 100",100);
  glutAddMenuEntry("+ 1000",1000);  
  glutAddMenuEntry("+ 10000",10000);
  glutAddMenuEntry("+ 100000",+100000);


  menu = glutCreateMenu(fmenu);
  glutAddMenuEntry("Load", MNU_LOAD);
  glutAddSubMenu("Delay", sub_delay);
  glutAddMenuEntry("Embrulho", MNU_EMBRULHO);
  glutAddSubMenu("Translao", sub_trans);
  glutAddSubMenu("Zoom", sub_zoom);
  glutAddMenuEntry("Sair", MNU_SAI);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  // Adicionando as teclas de atalho
  glutKeyboardFunc(keyboard);
  glutSpecialFunc(keyboard_special);
 
  zoom = 1.0; 
  glutDisplayFunc(display);
  
  glutMainLoop();
  return 0;
}


// Funo de display do openGL
void display(void)
{
  char *str;
  Ponto p;

  Ponto min,atual;
  Cor corPontos = branco;
  Cor corFecho;
  Cor corCompara;
  
  corFecho.R = 1.0;
  corFecho.G = 0.8;
  corFecho.B = 0.1;

  corCompara.R = 0.0;
  corCompara.G = 1.0;
  corCompara.B = 0.0;

  p.X = 0.0;
  p.Y = 0.0;

  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);
  
  glScaled(zoom,zoom,zoom);
  glTranslated(trans_x, trans_y, 0.0);

  marcaPontos(corPontos);
  
  zoom = 1.0;
  trans_x = 0.0;
  trans_y = 0.0;
  glutSwapBuffers();
}

// Funo de display da sub-janela que imprime caracteres na tela
void display_load() {  
  char *str;

  /* Clear the color buffer. */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glPushMatrix();
  glClearColor(0.0, 0.0, 0.0, 0.0);
  

  glTranslatef(-2.8, 2.6, 0.0);
  str = "Entre com o nome do arquivo:";
    
  glScalef(1 / 200.0, 1 / 200.0, 1 / 200.0);
  
  glPushMatrix();
  glColor3f(1.0, 1.0, 1.0);
  txfRenderString(txf, str, strlen(str));

  str = nome_arq;
  txfRenderString(txf, str, strlen(str));
  glPopMatrix();

  glPopMatrix();

  glutSwapBuffers();
  
}








